@@ -1,69 +1,24 @@
 /*
- * $Header: /home/cvs/jakarta-commons/jxpath/src/java/org/apache/commons/jxpath/util/ValueUtils.java,v 1.15 2003/03/11 00:59:34 dmitri Exp $
- * $Revision: 1.15 $
- * $Date: 2003/03/11 00:59:34 $
+ * Copyright 1999-2004 The Apache Software Foundation
  *
- * ====================================================================
- * The Apache Software License, Version 1.1
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
+ *     http://www.apache.org/licenses/LICENSE-2.0
  *
- * Copyright (c) 1999-2003 The Apache Software Foundation.  All rights
- * reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution, if
- *    any, must include the following acknowlegement:
- *       "This product includes software developed by the
- *        Apache Software Foundation (http://www.apache.org/)."
- *    Alternately, this acknowlegement may appear in the software itself,
- *    if and wherever such third-party acknowlegements normally appear.
- *
- * 4. The names "The Jakarta Project", "Commons", and "Apache Software
- *    Foundation" must not be used to endorse or promote products derived
- *    from this software without prior written permission. For written
- *    permission, please contact apache@apache.org.
- *
- * 5. Products derived from this software may not be called "Apache"
- *    nor may "Apache" appear in their names without prior written
- *    permission of the Apache Group.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Apache Software Foundation and was
- * originally based on software copyright (c) 2001, Plotnix, Inc,
- * <http://www.plotnix.com/>.
- * For more information on the Apache Software Foundation, please see
- * <http://www.apache.org/>.
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
 package org.apache.commons.jxpath.util;
 
 import java.beans.IndexedPropertyDescriptor;
 import java.beans.PropertyDescriptor;
 import java.lang.reflect.Array;
+import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
 import java.util.ArrayList;
@@ -82,10 +37,11 @@ import org.apache.commons.jxpath.JXPathException;
  * Collection and property access utilities.
  *
  * @author Dmitri Plotnikov
- * @version $Revision: 1.15 $ $Date: 2003/03/11 00:59:34 $
+ * @version $Revision: 1.19 $ $Date: 2004/04/04 22:06:36 $
  */
 public class ValueUtils {
     private static Map dynamicPropertyHandlerMap = new HashMap();
+    private static final int UNKNOWN_LENGTH_MAX_COUNT = 16000;
 
     /**
      * Returns true if the object is an array or a Collection
@@ -94,7 +50,8 @@ public class ValueUtils {
         if (value == null) {
             return false;
         }
-        else if (value.getClass().isArray()) {
+        value = getValue(value);
+        if (value.getClass().isArray()) {
             return true;
         }
         else if (value instanceof Collection) {
@@ -131,6 +88,42 @@ public class ValueUtils {
                 
         return 0;
     }
+    
+    /**
+     * If there is a regular non-indexed read method for this property,
+     * uses this method to obtain the collection and then returns its
+     * length.
+     * Otherwise, attempts to guess the length of the collection by
+     * calling the indexed get method repeatedly.  The method is supposed
+     * to throw an exception if the index is out of bounds. 
+     */
+    public static int getIndexedPropertyLength(
+        Object object,
+        IndexedPropertyDescriptor pd) 
+    {
+        if (pd.getReadMethod() != null) {
+            return getLength(getValue(object, pd));
+        }
+        
+        Method readMethod = pd.getIndexedReadMethod();
+        if (readMethod == null) {
+            throw new JXPathException(
+                "No indexed read method for property " + pd.getName());
+        }
+
+        for (int i = 0; i < UNKNOWN_LENGTH_MAX_COUNT; i++) {
+            try {
+                readMethod.invoke(object, new Object[] { new Integer(i)});
+            }
+            catch (Throwable t) {
+                return i;
+            }
+        }
+        
+        throw new JXPathException(
+            "Cannot determine the length of the indexed property "
+                + pd.getName());
+    }
 
     /**
      * Returns the length of the supplied collection. If the supplied object
@@ -140,7 +133,8 @@ public class ValueUtils {
         if (collection == null) {
             return 0;
         }
-        else if (collection.getClass().isArray()) {
+        collection = getValue(collection);
+        if (collection.getClass().isArray()) {
             return Array.getLength(collection);
         }
         else if (collection instanceof Collection) {
@@ -219,7 +213,7 @@ public class ValueUtils {
      * Returns the index'th element from the supplied collection.
      */
     public static Object remove(Object collection, int index) {
-        collection = openContainers(collection);
+        collection = getValue(collection);
         if (collection == null) {
             return null;
         }
@@ -277,7 +271,7 @@ public class ValueUtils {
      * Returns the index'th element of the supplied collection.
      */
     public static Object getValue(Object collection, int index) {
-        collection = openContainers(collection);
+        collection = getValue(collection);
         Object value = collection;
         if (collection != null) {
             if (collection.getClass().isArray()) {
@@ -314,7 +308,7 @@ public class ValueUtils {
      * Converts the value to the required type if necessary.
      */
     public static void setValue(Object collection, int index, Object value) {
-        collection = openContainers(collection);
+        collection = getValue(collection);
         if (collection != null) {
             if (collection.getClass().isArray()) {
                 Array.set(
@@ -422,8 +416,19 @@ public class ValueUtils {
                         bean,
                         new Object[] { new Integer(index)});
                 }
+            }            
+            catch (InvocationTargetException ex) {
+                Throwable t =
+                    ((InvocationTargetException) ex).getTargetException();
+                if (t instanceof ArrayIndexOutOfBoundsException) {
+                    return null;
+                }
+                
+                throw new JXPathException(
+                    "Cannot access property: " + propertyDescriptor.getName(),
+                    t);
             }
-            catch (Exception ex) {
+            catch (Throwable ex) {
                 throw new JXPathException(
                     "Cannot access property: " + propertyDescriptor.getName(),
                     ex);
@@ -461,7 +466,6 @@ public class ValueUtils {
                 }
             }
             catch (Exception ex) {
-                ex.printStackTrace();
                 throw new RuntimeException(
                     "Cannot access property: "
                         + propertyDescriptor.getName()
@@ -471,7 +475,7 @@ public class ValueUtils {
         }
         // We will fall through if there is no indexed read
         Object collection = getValue(bean, propertyDescriptor);
-        if (isCollection(openContainers(collection))) {
+        if (isCollection(collection)) {
             setValue(collection, index, value);
         }
         else if (index == 0) {
@@ -487,11 +491,11 @@ public class ValueUtils {
      * If the parameter is a container, opens the container and
      * return the contents.  The method is recursive.
      */
-    private static Object openContainers(Object collection) {
-        while (collection instanceof Container) {
-            collection = ((Container) collection).getValue();
+    public static Object getValue(Object object) {
+        while (object instanceof Container) {
+            object = ((Container) object).getValue();
         }
-        return collection;
+        return object;
     }
     
     /**
